/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2007,2008 Bull S.A.S.
 * Contact: cmi@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: AbsClusterViewManager.java 2414 2009-10-22 13:43:26Z eyindanga $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.controller.common;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;

import net.jcip.annotations.ThreadSafe;

import org.ow2.cmi.admin.CMIMBeanConfigException;
import org.ow2.cmi.admin.MBeanUtils;
import org.ow2.cmi.controller.client.ClientClusterViewManager;
import org.ow2.cmi.controller.factory.ClusterViewManagerFactory;
import org.ow2.cmi.event.CMIEventComponent;
import org.ow2.cmi.ha.SessionId;
import org.ow2.cmi.lb.policy.IPolicy;
import org.ow2.cmi.lb.util.PolicyFactory;
import org.ow2.cmi.lb.util.PolicyFactoryException;
import org.ow2.cmi.reference.CMIReference;
import org.ow2.cmi.reference.CMIReferenceable;
import org.ow2.cmi.reference.ObjectNotFoundException;
import org.ow2.util.component.ComponentManager;
import org.ow2.util.component.api.Component;
import org.ow2.util.component.api.ComponentException;
import org.ow2.util.component.api.IComponentManager;
import org.ow2.util.event.api.IEventDispatcher;
import org.ow2.util.event.impl.EventDispatcher;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.api.Pool;

/**
 * Common abstraction of a manager of the cluster view.
 * @author The new CMI team
 */
@ThreadSafe
public abstract class AbsClusterViewManager implements ClusterViewManager {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory.getLog(AbsClusterViewManager.class);

    /**
     * State of the manager.
     */
    private State state = State.STOPPED;

    /**
     * Map local data for each clustered RMI object.
     */
    private final ConcurrentHashMap<String, LocalObjectData> localObjectData =
        new ConcurrentHashMap<String, LocalObjectData>();

    /**
     * Objects that require an update of local data when its global data change.
     */
    private final Set<String> watchedObjects =
        Collections.synchronizedSet(new HashSet<String>());

    /**
     * A factory of daemon.
     */
    private final CMIThreadFactory cmiThreadFactory;

    /**
     * Universally Unique Identifier.
     */
    private final UUID uuid;

    /**
     * The last identifier of session.
     */
    private final AtomicLong sessionNb = new AtomicLong(0);

    /**
     * The event dispatcher.
     */
    private final IEventDispatcher dispatcher;

    /**
     * Manager of components.
     */
    private IComponentManager componentManager;

    /**
     * The provider id.
     */
    private final String cmiManagedObjectId;

    /**
     * True if the manager is destroyed (i.e. it cannot any more start).
     */
    private boolean destroyed;


    /**
     * Creates a new instance of AbsClusterViewManager with initializing a factory of daemon.
     */
    protected AbsClusterViewManager() {
        String managerName;
        if(this instanceof ClientClusterViewManager) {
            managerName = "ClientClusterViewManager";
        } else {
            managerName = "ServerClusterViewManager";
        }
        cmiThreadFactory = new CMIThreadFactory(managerName);

        // Identify this manager
        uuid = UUID.randomUUID();

        //Create the event dispatcher
        this.dispatcher = new EventDispatcher();

        cmiManagedObjectId = "/cmi";
    }

    public IEventDispatcher getEventDispatcher() {
        return dispatcher;
    }

    /**
     * Gets component with the given name.
     * @param componentName the name of the component.
     * @return the component (if any)
     */
    public final Component getComponent(final String componentName) {
        // ask registry if present.
        if (this.componentManager != null) {
            return this.componentManager.getComponentRegistry().getComponent(componentName);
        }
        return null;
    }

    /**
     * Get a reference to the first component matching the interface.
     * @param <T> The interface type.
     * @param itf The interface class.
     * @return The component.
     */
    public <T extends Component> T getComponent(final Class<T> itf) {
        try {
            return this.componentManager.getComponentRegistry().getComponents(itf).get(0);
        } catch (IndexOutOfBoundsException e) {
            LOGGER.debug("No component", e);
            return null;
        }
    }

    /**
     * @return Returns the ComponentManager used by this instance.
     */
    public final IComponentManager getComponentManager() {
        return this.componentManager;
    }


    /* (non-Javadoc)
     * @see org.ow2.cmi.CMIManagedObject#getCMIManagedObjectId()
     */
    public String getCMIManagedObjectId() {
        return cmiManagedObjectId;
    }

    public final synchronized State getState() {
        return state;
    }

    protected final void setState(final State state) {
        this.state = state;
    }

    /**
     * Returns the instance of LB policy associated with the given name of object.
     * @param objectName a name of object
     * @return the instance of LB policy associated with the given name of object
     */
    public final IPolicy<CMIReference> getPolicy(final String objectName) {
        localObjectData.putIfAbsent(objectName, new LocalObjectData(objectName));
        return localObjectData.get(objectName).getPolicy();
    }

    /**
     * Associates a name of of object with an instance of LB policy.
     * @param objectName a name of object
     * @param lbPolicy an instance of LB policy to associate with the given object
     */
    private void setPolicy(final String objectName, final IPolicy<CMIReference> lbPolicy) {
        localObjectData.putIfAbsent(objectName, new LocalObjectData(objectName));
        localObjectData.get(objectName).setPolicy(lbPolicy);
    }

    /**
     * Returns true if the object with the given name has a pool of CMIReferenceable.
     * @param objectName a name of object
     * @return true if the object with the given name has a pool of CMIReferenceable
     */
    protected final boolean hasPool(final String objectName) {
        localObjectData.putIfAbsent(objectName, new LocalObjectData(objectName));
        return localObjectData.get(objectName).getPool() != null;
    }

    /**
     * Returns a pool of CMIReferenceable for an object with the given name.
     * @param objectName a name of object
     * @return a pool of CMIReferenceable for an object with the given name
     */
    public final Pool<CMIReferenceable<?>, CMIReference> getPool(final String objectName) {
        localObjectData.putIfAbsent(objectName, new LocalObjectData(objectName));
        return localObjectData.get(objectName).getPool();
    }

    /**
     * Associates a name of object with a pool of CMIReferenceable.
     * @param objectName a name of object
     * @param pool a pool of CMIReferenceable
     */
    public final void setPool(final String objectName, final Pool<CMIReferenceable<?>, CMIReference> pool) {
        localObjectData.putIfAbsent(objectName, new LocalObjectData(objectName));
        localObjectData.get(objectName).setPool(pool);
    }

    /**
     * Updates a policy of load-balancing for a object with the given name.
     * @param objectName a name of object
     * @throws ObjectNotFoundException if none object has the given name
     * @throws PolicyFactoryException if the policy cannot be updated
     */
    protected final void updatePolicy(final String objectName) throws ObjectNotFoundException, PolicyFactoryException {
        LOGGER.debug("Updating policy for {0}", objectName);
        PolicyFactory<CMIReference> policyFactory = new PolicyFactory<CMIReference>(this);
        IPolicy<CMIReference> policy = policyFactory.getPolicy(objectName);
        setPolicy(objectName, policy);
    }

    /**
     * Returns true if the local data for the object with the given name has been initialized.
     * @param objectName a name of object
     * @return true if the local data for the object with the given name has been initialized
     */
    protected final boolean isWatched(final String objectName) {
        return watchedObjects.contains(objectName);
    }

    protected final void watch(final String objectName) {
        watchedObjects.add(objectName);
        LOGGER.debug("Object {0} is now watched", objectName);
    }

    /**
     * Returns the set of names of object that has been initialized.
     * @return the set of names of object that has been initialized
     */
    protected final Set<String> getNamesOfWatchedObject() {
        return new HashSet<String>(watchedObjects);
    }

    /**
     * Returns the factory to create CMI daemons.
     * @return the factory to create CMI daemons
     */
    protected final CMIThreadFactory getCmiThreadFactory() {
        return cmiThreadFactory;
    }

    /**
     * @return the UUID for this manager
     */
    public final UUID getUUID() {
        return uuid;
    }

    /**
     * @return a new identifier of session
     */
    public final SessionId getSessionId() {
        return new SessionId(uuid,
                sessionNb.incrementAndGet());
    }

    /**
     * @return the configuration used by this manager
     */
    public final IConfig getConfig() {
        return ClusterViewManagerFactory.getFactory().getConfig();
    }

    public final synchronized boolean start() throws ClusterViewManagerException {
        if(destroyed) {
            throw new ClusterViewManagerException(
                    "This instance of cluster view manager can no more be used !!!");
        }
        if(getConfig() == null) {
            throw new ClusterViewManagerException(
                "This instance of cluster view manager has no configuration !!!");
        }
        if(getState().equals(State.STOPPED)) {
            setState(State.STARTING);

            // Create component manager around components
            this.componentManager = new ComponentManager(getConfig().getComponents());

            if (getConfig().isAutoConfigureComponents()) {
                // Init components
                try {
                    getComponentManager().initComponents();
                } catch (ComponentException e) {
                    throw new ClusterViewManagerException("Cannot init components", e);
                }

                // Start components
                try {
                    getComponentManager().startComponents();
                } catch (ComponentException e) {
                    throw new ClusterViewManagerException("Cannot start components", e);
                }
            }

            // Start the event dispatcher
            getEventDispatcher().start();

            // Register to event component.
            CMIEventComponent eventComponent = getComponent(CMIEventComponent.class);
            if (eventComponent != null) {
                eventComponent.registerCMIManagedObject(this, getEventDispatcher());
            }

            if(getConfig().isAdminEnabled()) {
                // Registers CMIMBean
                try {
                    MBeanUtils.registerCMIMBean(this);
                } catch (CMIMBeanConfigException e) {
                    LOGGER.warn("Cannot registers the CMI MBean", e);
                }
            }
            return starting();
        }
        return getState().equals(ClusterViewManager.State.AVAILABLE);
    }

    public synchronized final void stop() {
        if(getState().equals(State.STARTED) || getState().equals(State.AVAILABLE)) {
            setState(State.STOPPING);
            stopping();
            if(getConfig().isAdminEnabled()) {
                // Unregister the CMI MBean
                try {
                    MBeanUtils.unregisterCMIMBean();
                } catch (CMIMBeanConfigException e) {
                    LOGGER.error("Cannot unregister CMIMBean", e);
                }
            }
            setState(State.STOPPED);
            LOGGER.info("Cluster view manager stopped.");
        }
    }

    protected abstract boolean starting() throws ClusterViewManagerException;

    protected abstract void stopping();

    public void destroy() {
        stop();
        destroyed = true;
    }

}
